#ifndef Fp_cc 
#define Fp_cc

#include <iostream.h>
#include <builtin.h>
#include "Fp.h"
#include <Rational.h>

extern int len(int x);

/*******************************************************************
constructor
*******************************************************************/

Fp::Fp (Fp &x) {
	q = x.q;
};

Fp::Fp (int x) { // there is no test for x !!!
	q = x;
};

int Fp::prime = 31991;

/*******************************************************************
ops
*******************************************************************/

Fp& Fp::operator = (Fp &x) {
	q = x.q;
	return *this;
};

Fp& Fp::operator = (int y) {
	q = ((long)y + prime) % prime;
	return *this;
};

int Fp::operator != (Fp &y) {
	return (q != y.q);
};

int Fp::operator == (Fp &y) {
	return (q == y.q);
};

int Fp::operator != (int y) {
	return (q != y);
};

int Fp::operator == (int y) {
	return (q == y);
};

/*******************************************************************
unary ops
*******************************************************************/

Fp& operator + (Fp &x) {
	return x;
};

Fp  operator - (Fp &x) {
	return Fp(x.prime - x.q);
};

/*******************************************************************
binary ops
*******************************************************************/

Fp Fp::operator + (Fp &y) {
	return Fp(((long)q + y.q) % prime);
};

Fp Fp::operator - (Fp &y) {
	return Fp((q - y.q + (long)prime) % prime);
};

Fp Fp::operator * (Fp &y)	{
	return Fp(((long)q * y.q) % prime);
};

Fp Fp::operator / (Fp y) {
	y.invert();
	return (((long)q * y.q) % prime);
};

Fp& Fp::operator += (Fp &y) {
	q = ((long)q + y.q) % prime;
	return *this;
};

Fp& Fp::operator -= (Fp &y) {
	q = (q - y.q + (long)prime) % prime;
	return *this;
};

Fp& Fp::operator *= (Fp &y) {
	q = ((long)q * y.q) % prime;
	return *this;
};

Fp& Fp::operator /= (Fp y) {
   y.invert();
	q = ((long)q * y.q) % prime;
	return *this;
};

void Fp::invert() {
	if (q == 0) {
		cout << "\nFp: division by zero\n";
		exit(1);
	};
	if (q == 1)
		return;
	int r = 0;
	long tmp = 0;
	long old = 0;
	long act = 1;
	int x = prime;
	do {
		r = x % q;
		tmp = act;
		act = old - (x / q) * act;
		old = tmp;
		x = q;
		q = r;
	} while (r != 1);
	q = (act + prime) % prime;
};

/*******************************************************************
I/O
*******************************************************************/

ostream& operator << (ostream &s, Fp &x) {
	if (x.q > x.prime / 2)
		return s << x.q - x.prime;
	else
		return s << x.q;
};

istream& operator >> (istream &s, Fp &x) {
	Rational tmp;
	s >> dec >> tmp;
	Integer res = tmp.denominator();
	res %= Fp::prime;
	x.q = atoi(Itoa(res));
	x.invert();
	res = tmp.numerator();
	res %= Fp::prime;
	Fp y;
	y.q = atoi(Itoa(res));
	x *= y;
	return s;
};

/*******************************************************************
common
*******************************************************************/

int len(Fp &x) { //
	if (x.q > x.prime / 2)
		return len(x.q - x.prime);
	else
		return len(x.q);
};

Fp pow(Fp &x, int n) { // there is no check for n<0
	switch (n) {
	case 0:
		return Fp(1);
	case 1:
		return x;
	case 2:
		return x * x;
	case 3:
		return x * x * x;
	default:
		Fp y(1), z = x;
		for(;;) {
			if (odd(n)) {// n odd
				n /= 2;
				y *= z;
				if (! n)
					return y;
				z *= z;
			} 					// n even
			else {
				n /= 2;
				z *= z;
			};
		};
	};
};

Fp abs(Fp x) {
	if (sign(x) < 0)
		x.q = x.prime - x.q;
	return x;
};

int sign(Fp &x) { //
	if (x.q == 0)
		return 0;
	if (x.q > x.prime / 2)
		return -1;
	else
		return 1;
};


#endif
