#ifndef Fp8_cc
#define Fp8_cc

#include <iostream.h>
#include <builtin.h>
#include "Fp.h"
#include "Fp8.h"

extern int len(int x);

/*******************************************************************
constructor
*******************************************************************/

Fp8::Fp8 (Fp8 &x) {
	q = x.q;
};

Fp8::Fp8 (int x) { // there is no test for x !!!
	q = (unsigned char)x;
};

int Fp8::prime = 101;

unsigned char Fp8::add[256][256];
unsigned char Fp8::sub[256][256];
unsigned char Fp8::mul[256][256];
unsigned char Fp8::div[256][256];

/*******************************************************************
ops
*******************************************************************/

Fp8& Fp8::operator = (Fp8 &x) {
	q = x.q;
	return *this;
};

Fp8& Fp8::operator = (int y) {
	q = y % prime;
	return *this;
};

int Fp8::operator != (Fp8 &y) {
	return (q != y.q);
};

int Fp8::operator == (Fp8 &y) {
	return (q == y.q);
};

int Fp8::operator != (int y) {
	return (q != y);
};

int Fp8::operator == (int y) {
	return (q == y);
};

/*******************************************************************
unary ops
*******************************************************************/

Fp8& operator + (Fp8 &x) {
	return x;
};

Fp8  operator - (Fp8 &x) {
	return Fp8(Fp8::sub[0][x.q]);
};

/*******************************************************************
binary ops
*******************************************************************/

Fp8 Fp8::operator + (Fp8 &y) {
	return Fp8(add[q][y.q]);
};

Fp8 Fp8::operator - (Fp8 &y) {
	return Fp8(sub[q][y.q]);
};

Fp8 Fp8::operator * (Fp8 &y)  {
	return Fp8(mul[q][y.q]);
};

Fp8 Fp8::operator / (Fp8 y) {
	return Fp8(div[q][y.q]);
};

Fp8& Fp8::operator += (Fp8 &y) {
	q = add[q][y.q];
	return *this;
};

Fp8& Fp8::operator -= (Fp8 &y) {
	q = sub[q][y.q];
	return *this;
};

Fp8& Fp8::operator *= (Fp8 &y) {
	q = mul[q][y.q];
	return *this;
};

Fp8& Fp8::operator /= (Fp8 y) {
	q = div[q][y.q];
	return *this;
};

void Fp8::invert() {
	if (q == 0) {
		cout << "\nFp8: division by zero\n";
		exit(1);
	};
	q = div[1][q];
};

/*******************************************************************
I/O
*******************************************************************/

ostream& operator << (ostream &s, Fp8 &x) {
	return s << Fp(x.q);
};

istream& operator >> (istream &s, Fp8 &x) {
	Fp tmp;
	s >> tmp;
	x.q = tmp.q % Fp8::prime;
	return s;
};

/*******************************************************************
common
*******************************************************************/

int len(Fp8 &x) { //
	return len(Fp(x.q));
};

Fp8 abs(Fp8 x) {
	if (sign(x) < 0)
		x.q = x.sub[0][x.q];
	return x;
};

int sign(Fp8 &x) { //
	if (x.q == 0)
		return 0;
	if (x.q > x.prime / 2)
		return -1;
	else
		return 1;
};

void Fp8::init() {
	Fp::prime = prime;
	int i,j;
	for(i = 0; i < prime; i++)
		for(j = 0; j < prime; j++) {
			add[i][j] = (unsigned char) ((int(i)+j) % prime);
			sub[i][j] = (unsigned char) ((int(i)-j+prime) % prime);
			mul[i][j] = (unsigned char) ((int(i)*j) % prime);
		};
	for(i = 0; i < prime; i++)
		for(j = 1; j < prime; j++) {
			Fp tmp1(i);
			Fp tmp2(j);
			tmp1 /= tmp2;
			div[i][j] = (unsigned char) tmp1.q;
		};
};

#endif
