#ifndef QN_cc
#define QN_cc


#include <iostream.h>
#include <builtin.h>
#include "QN.h"

int is_el_Q(QN &x) {
	for(int i = 1; i < QN::n; i++)
		if (x.a[i] != 0) return 0;
	return 1;
};

template <class K>
void swap(K &a, K &b) {
	K tmp = b;
	b = a;
	a = tmp;
	return;
};

extern int len(int x);
extern int len(Rational x);

int QN::n = 1;
int QN::N = 1;
Rational *(*QN::mat) = new Rational*[1];
QN *QN_phi;
int *QN::r = new int[QN::n+1];

/*******************************************************************
constructor
*******************************************************************/

QN::QN (QN &x) {
	a = new Rational[n];
	for(int i = 0; i < n; i++)
		a[i] = x.a[i];
};

QN::QN (int x) {
	a = new Rational[n];
	for(int i = 1; i < n; i++)
		a[i] = 0;
	a[0] = x;
};

QN::QN (Rational &x) {
	a = new Rational[n];
	for(int i = 1; i < n; i++)
		a[i] = 0;
	a[0] = x;
};

QN::~QN() {
	delete[] a;
};

/*******************************************************************
ops
*******************************************************************/

QN& QN::operator = (QN &x) {
	for(int i = 0; i < n; i++)
		a[i] = x.a[i];
	return *this;
};

QN& QN::operator = (Rational &x) {
	for(int i = 1; i < n; i++)
		a[i] = 0;
	a[0] = x;
	return *this;
};

QN& QN::operator = (int y) {
	for(int i = 1; i < n; i++)
		a[i] = 0;
	a[0] = y;
	return *this;
};

int QN::operator != (QN &y) {
	for(int i = 0; i < n; i++)
		if(a[i] != y.a[i]) return 1;
	return 0;
};

int QN::operator == (QN &y) {
	for(int i = 0; i < n; i++)
		if(a[i] != y.a[i]) return 0;
	return 1;
};

int QN::operator != (int y) {
	if (a[0] != y) return 1;
	for(int i = 1; i < n; i++)
		if(sign(a[i]) != 0) return 1;
	return 0;
};

int QN::operator == (int y) {
	if (a[0] != y) return 0;
	for(int i = 1; i < n; i++)
		if(sign(a[i]) != 0) return 0;
	return 1;
};

/*******************************************************************
unary ops
*******************************************************************/

QN& operator + (QN &x) {
	return x;
};

QN operator - (QN x) {
	for(int i = 0; i < QN::n; i++)
		x.a[i].negate();
	return x;
};

/*******************************************************************
binary ops
*******************************************************************/

QN QN::operator + (QN &y) {
	QN tmp(*this);
	for(int i = 0; i < n; i++)
		tmp.a[i] += y.a[i];
	return tmp;
};

QN QN::operator - (QN &y) {
	QN tmp(*this);
	for(int i = 0; i < n; i++)
		tmp.a[i] -= y.a[i];
	return tmp;
};

QN QN::operator * (Rational &y) {
	QN tmp(*this);
	for(int i = 0; i < n; i++)
		tmp.a[i] *= y;
	return tmp;
};

QN QN::operator * (QN &y)	{
	if(is_el_Q(y)) {
		QN tmp(*this);
		for(int k = 0; k < n; k++)
			tmp.a[k] *= y.a[0];
		return tmp;
	};
	QN tmp(0);
	int i,j;
	for(i = 0; i < n; i++)
		for(j = 0; j < n; j++)
			if(i+j < n)
				tmp.a[i+j] += a[i] * y.a[j];
			else
				if (i+j >= N)
					tmp.a[i+j-N] += a[i] * y.a[j];
				else
					tmp += QN_phi[i+j-n] * a[i] * y.a[j];
	return tmp;
};

void QN::invert();
QN& QN::operator *= (QN &y);

QN QN::operator / (QN y) {
	QN tmp(*this);
	y.invert();
	return tmp *= y;
};

QN& QN::operator += (QN &y) {
	for(int i = 0; i < n; i++)
		a[i] += y.a[i];
	return *this;
};

QN& QN::operator -= (QN &y) {
	for(int i = 0; i < n; i++)
		a[i] -= y.a[i];
	return *this;
};

QN& QN::operator *= (Rational &y) {
	if(is_el_Q(*this)) {
		a[0] *= y;
		return *this;
	};
	for(int i = 0; i < n; i++)
		a[i] *= y;
	return *this;
};

QN& QN::operator *= (QN &y) {
	if(is_el_Q(y)) return *this *= y.a[0];
	QN tmp(*this);
	*this = 0;
	int i,j;
	for(i = 0; i < n; i++)
		for(j = 0; j < n; j++)
			if(i+j < n)
				a[i+j] += tmp.a[i] * y.a[j];
			else
				if (i+j >= N)
					a[i+j-N] += tmp.a[i] * y.a[j];
				else
					*this += QN_phi[i+j-n] * tmp.a[i] * y.a[j];
	return *this;
};

QN& QN::operator /= (QN y) {
	y.invert();
	return *this *= y;
};

void QN::invert() {
	if (sign(*this) == 0) {
		cerr << "QN : division by zero !!! \n";
		exit(3);
	};
	if(is_el_Q(*this)) {
		a[0].invert();
		return;
	};
	int i,j,k;

	for(i = 0; i < n; i++)
		for(j = 0; j <= n; j++)
			mat[i][j] = 0;
	mat[0][n] = 1;

	for(i = 0; i <= n; i++) r[i] = i;

	for(i = 0; i < n; i++) {
		if(sign(a[i]) == 0) continue;
		for(j = 0; j < n; j++)
			if(i+j < n)
				mat[i+j][j] += a[i];
			else
				if(i+j >= N)
					mat[i+j-N][j] += a[i];
				else
					for(k = 0; k < n; k++)
						mat[k][j] += a[i] * QN_phi[i+j-n].a[k];
	};

	for (i = 0; i < n; i++) {
		if (sign(mat[r[i]][i]) == 0) {
			for(j = i + 1; j < n; j++) {
				if (sign(mat[r[j]][i]) == 0)
					continue;
				swap(r[i], r[j]);
				break;
			};
		};

		for(j = i + 1; j < n + 1; j++)
			mat[r[i]][j] /= mat[r[i]][i];
		mat[r[i]][i] = 1;

		for (j = i + 1; j < n; j++) {	// elim col i
			if(mat[r[j]][i] != 0) {
				for (k = i + 1; k < n + 1; k++)
					mat[r[j]][k] -= mat[r[i]][k] * mat[r[j]][i];
				mat[r[j]][i] = 0;
			};
		};
	};
	// matrix has now 'triangular' shape
	for(i = n - 2; i >= 0; i--) {	// which row to eliminate
		for(j = i + 1; j < n; j++) {	// col to handle
			mat[r[i]][n] -= mat[r[i]][j] * mat[r[j]][n];
			mat[r[i]][j] = 0;
		};
	};

	for(i = 0; i < n; i++)
		a[i] = mat[r[i]][n];
};

/*******************************************************************
I/O
*******************************************************************/

ostream& operator << (ostream &s, QN &x) {
	if(is_el_Q(x)) return s << x.a[0];
	s << "(";
	for(int i = 0; i < QN::n - 1; i++)
		s << x.a[i] << ",";
	s << x.a[QN::n - 1] << ")";
	return s;
};

istream& operator >> (istream &s, QN &x) {
	char c;
	for(int i = 0; i < x.n-1; i++) {
		x.a[i] = 0;
		s >> dec >> x.a[i];
		s.get(c);
	};
	x.a[x.n-1] = 0;
	s >> dec >> x.a[x.n-1];
	return s;
};

/*******************************************************************
common
*******************************************************************/

int len(QN &x) { //
	if(is_el_Q(x)) return len(x.a[0]);
	int tmp = QN::n + 1;
	for(int i = 0; i < QN::n; i++) tmp += len(x.a[i]);
	return tmp;
};

QN abs(QN &x) {
	if (is_el_Q(x)) return abs(x.a[0]);
	return x;
};

int sign(QN &x) {
	if (is_el_Q(x)) return (sign(x.a[0]));
	return 1;
};

int is_prime(int n) {
	for(int i = 2; i <= n/2; i++)
		if(n % i == 0) return 0;
	return 1;
};

int mu(int n) {
	if(n == 1) return 1;
	if(n == 2) return -1;
	if(n == 3) return -1;
	if(n == 4) return 0;

	int m = n;
	int r = 0;
	for(int i = 2; i <= n/2; i++) {
		if(is_prime(i) == 0) continue;
		if(m % i == 0) {
			r++;
			m /= i;
			if(m % i == 0) return 0;
		};
	};
	if(r == 0) return -1;
	if(r % 2 == 0) return 1;
	return -1;
};

void set(int n) {
	int i,j,k,m;
	int max = 1;

	for(i = 1; i <= n; i++) {
		if(n % i != 0) continue;
		if(mu(i) > 0) max += n/i;
	};
	Rational *a = new Rational[max];
	Rational *b = new Rational[max];
	Rational *c = new Rational[max];
	Rational *d = new Rational[max];
	for(i = 0; i < max; i++) {
		a[i] = 0;
		b[i] = 0;
		c[i] = 0;
		d[i] = 0;
	};
	a[n] = 1;
	a[0] = -1;
	m = n;

	// numerator -> a
	for(i = 2; i <= n; i++) {
		if(n % i != 0) continue;
		if(mu(i) <= 0) continue;
		b[n/i] = 1;
		b[0] = -1;
		// m = len a
		for(j = 0; j <= m; j++)
			for(k = 0; k <= n/i; k++)
				c[j+k] += a[j] * b[k];
		m += n/i;
		for(j = 0; j < max; j++) {
			a[j] = c[j];
			b[j] = 0;
			c[j] = 0;
		};
	};

	b[0] = 1;
	m = 0;
	// denominator -> b
	for(i = 2; i <= n; i++) {
		if(n % i != 0) continue;
		if(mu(i) >= 0) continue;
		c[n/i] = 1;
		c[0] = -1;
		// m = len b
		for(j = 0; j <= m; j++)
			for(k = 0; k <= n/i; k++)
				d[j+k] += b[j] * c[k];
		m += n/i;
		for(j = 0; j < max; j++) {
			b[j] = d[j];
			c[j] = 0;
			d[j] = 0;
		};
	};

	// a / b -> c
	for(i = max - 1; i >= m; i--) {
		if(sign(a[i]) == 0) continue;
		c[i-m] = a[i] / b[m];
		for(k = 0; k <= m; k++)
			a[i-m+k] -= b[k] * c[i-m];
	};

	for(i = 0; i < QN::n; i++)
		delete[] QN::mat[i];
	delete[] QN::mat;
	delete[] QN::r;

	QN::N = n;
	QN::n = max - m - 1;
	QN::r = new int[QN::n+1];
	QN::mat = new Rational*[QN::n];
	for(i = 0; i < QN::n; i++)
		QN::mat[i] = new Rational[QN::n+1];

	m = QN::N - QN::n;

	delete[] QN_phi;
	QN_phi = new QN[m];

	for(i = 0; i < QN::n; i++)
		QN_phi[0].a[i] = - c[i];

	for(i = 1; i < m; i++) {
		for(j = 1; j < QN::n; j++)
			QN_phi[i].a[j] = QN_phi[i-1].a[j-1];
		QN_phi[i].a[0] = 0;
		QN_phi[i] += QN_phi[0] * QN_phi[i-1].a[QN::n-1];
	};

	delete[] a;
	delete[] b;
	delete[] c;
	delete[] d;
};

#endif
