#include <iostream.h>
#include <builtin.h>
#include "monom.h"
#include <minmax.h>

extern int dim;
extern int proj;
extern char varstring[];
extern int len(int x);

/*******************************************************************
constructors
*******************************************************************/

monom::monom(monom &x) {
	dim = x.dim;
	exp = new char[proj];
	for (int i = 0; i < dim; i++)
		exp[i] = x.exp[i];
};

monom::monom(char x) {
	dim = ::dim;
	exp = new char[proj];
	for (int i = 0; i < proj; i++)
		exp[i] = x;
};

/*******************************************************************
destructor
*******************************************************************/

monom::~monom () {
	delete[] exp;
};

/*******************************************************************
assign ops
*******************************************************************/

monom& monom::operator = (monom &x) {
	if (this != &x) {  //  x = x ?
		dim = x.dim;
		for (int i = 0; i < x.dim; i++)
			exp[i] = x.exp[i];
	};
	return *this;
};

monom& monom::operator = (char x) {
	for (int i = 0; i < proj; i++)
		exp[i] = x;
	return *this;
};

int monom::vars() {
	int tmp = 0;
	for (int i = 0; i < dim; i++) {
		if(exp[i] > 0)
			tmp++;
	};
	return tmp;
};

/*******************************************************************
relops
*******************************************************************/

extern int (*monom_ordering_lt)(monom &x, monom &y);
extern int (*monom_ordering_le)(monom &x, monom &y);
extern int (*proj_monom_ordering_lt)(monom &x, monom &y);
extern int (*proj_monom_ordering_le)(monom &x, monom &y);

int monom::operator < (monom &y) {
	if(dim != proj || y.dim != proj)
		return monom_ordering_lt(*this, y);
	else
		return proj_monom_ordering_lt(*this, y);
};

int monom::operator > (monom &y) {
	if(dim != proj || y.dim != proj)
		return monom_ordering_lt(y, *this);
	else
		return proj_monom_ordering_lt(y, *this);
};

int monom::operator <= (monom &y) {
	if(dim != proj || y.dim != proj)
		return monom_ordering_le(*this, y);
	else
		return proj_monom_ordering_le(*this, y);
};

int monom::operator >= (monom &y) {
	if(dim != proj || y.dim != proj)
		return monom_ordering_le(y, *this);
	else
		return proj_monom_ordering_le(y, *this);
};

int monom::operator == (monom &y) {
	for(int i = 0; i < min(dim, y.dim); i++) {
		if (exp[i] != y.exp[i])
			return 0;
	};
	return 1;
};

int aff_eq(monom &x, monom &y) {
	for(int i = 0; i < proj - 1; i++) {
		if (x.exp[i] != y.exp[i])
			return 0;
	};
	return 1;
};

int monom::operator != (monom &y) {
	for(int i = 0; i < min(dim, y.dim); i++) {
		if (exp[i] != y.exp[i])
			return 1;
	};
	return 0;
};

int lex_lt (monom &x, monom &y) {
	for(int i = 0; i < proj - 1; i++) {
		if (x.exp[i] != y.exp[i])
			return (x.exp[i] < y.exp[i]);
	};
	return 0;
};

int lex_le (monom &x, monom &y) {
	for(int i = 0; i < proj - 1; i++) {
		if (x.exp[i] != y.exp[i])
			return (x.exp[i] < y.exp[i]);
	};
	return 1;
};

int revlex_lt (monom &x, monom &y) {
	for(int i = proj - 2; i >= 0; i--) {
		if (x.exp[i] != y.exp[i])
			return (x.exp[i] > y.exp[i]);
	};
	return 0;
};

int revlex_le (monom &x, monom &y) {
	for(int i = proj - 2; i >= 0; i--) {
		if (x.exp[i] != y.exp[i])
			return (x.exp[i] > y.exp[i]);
	};
	return 1;
};

int deglex_le (monom &x, monom &y) {
	return (x.tot_deg() != y.tot_deg()) ?
		(x.tot_deg() < y.tot_deg()) : lex_le(x, y);
};

int deglex_lt (monom &x, monom &y) {
	return (x.tot_deg() != y.tot_deg()) ?
		(x.tot_deg() < y.tot_deg()) : lex_lt(x, y);
};

int degrevlex_le (monom &x, monom &y) {
	return (x.tot_deg() != y.tot_deg()) ?
		(x.tot_deg() < y.tot_deg()) : revlex_le(x, y);
};

int degrevlex_lt (monom &x, monom &y) {
	return (x.tot_deg() != y.tot_deg()) ?
		(x.tot_deg() < y.tot_deg()) : revlex_lt(x, y);
};

int Zn_order_lt (monom &x, monom &y) {
	int l = 0;
	for(int i = 0; i < x.dim; i++) {
		if (x.exp[i] > y.exp[i])
			return 0;
		if (x.exp[i] < y.exp[i])
			l++;
	};
	return (l > 0);
};

int Zn_order_le (monom &x, monom &y) {
	for(int i = 0; i < x.dim; i++)
		if (x.exp[i] > y.exp[i])
			return 0;
	return 1;
};

int monom::tot_deg() {
	int tmp = 0;
	for(int i = 0; i < proj - 1;	tmp += exp[i++]);
	return tmp;
};

void monom::put(ostream &s) {
	if (sign(*this) == 0)
		s << '1';
	else {
		for (int i = 0; i < dim; i++)
			if (exp[i]) {
				s << varstring[i];
				if (exp[i] != 1)
					s << (int)exp[i];
			};
	};
};

/*******************************************************************
the projective orderings
*******************************************************************/

int proj_deglex_le (monom &x, monom &y) {
	if (deglex_lt(x,y))
		return 1;
	else if (aff_eq(x,y))
		return (x.exp[proj - 1] <= y.exp[proj - 1]);
	return 0;
};

int proj_deglex_lt (monom &x, monom &y) {
	if (deglex_lt(x,y))
		return 1;
	else if (aff_eq(x,y))
		return (x.exp[proj - 1] < y.exp[proj - 1]);
	return 0;
};

int proj_degrevlex_le (monom &x, monom &y) {
	if (degrevlex_lt(x,y))
		return 1;
	else if (aff_eq(x,y))
		return (x.exp[proj - 1] <= y.exp[proj - 1]);
	return 0;
};

int proj_degrevlex_lt (monom &x, monom &y) {
	if (degrevlex_lt(x,y))
		return 1;
	else if (aff_eq(x,y))
		return (x.exp[proj - 1] < y.exp[proj - 1]);
	return 0;
};

/*******************************************************************
binary ops
*******************************************************************/

monom monom::operator * (monom &y) {
	monom tmp(0);
	for(int i = 0; i < min(dim, y.dim); i++)
		tmp.exp[i] = exp[i] + y.exp[i];
	return tmp;
};

monom monom::operator / (monom &y) { // there is no check for
	monom tmp(0);                                   // x > y
	for(int i = 0; i < min(dim, y.dim); i++)
		tmp.exp[i] = exp[i] - y.exp[i];
	return tmp;
};

monom& monom::operator *= (monom &y) {
	for(int i = 0; i < min(dim, y.dim); i++)
		exp[i] += y.exp[i];
	return *this;
};

monom& monom::operator /= (monom &y) {
	for(int i = 0; i < min(dim, y.dim); i++)
		exp[i] -= y.exp[i];
	return *this;
};

/*******************************************************************
I/O
*******************************************************************/

int sign(monom &x);

ostream& operator << (ostream &s, monom &x) {
	if (sign(x) == 0)
		s << '1';
	else {
		for (int i = 0; i < x.dim; i++)
			if (x.exp[i]) {
				s << varstring[i];
				if (x.exp[i] != 1)
					s << "^" << (int)x.exp[i];
			};
	};
	return s;
};

/*******************************************************************
common
*******************************************************************/

int len(monom &m) {
	int tmp = 0;
	for (int i = 0; i < m.dim; i++) {
		if(m.exp[i] == 1)
			tmp++;
		if(m.exp[i] > 1)
			tmp += len(m.exp[i]) + 2;
	};
	return tmp;
};

int sign(monom &x) {
	for(int i = 0; i < x.dim; i++)
		if (x.exp[i])
			return 1;
	return 0;
};

monom lcm(monom &x, monom &y) {
	monom tmp(x);
	tmp.dim = min(x.dim, y.dim);
	for(int i = 0; i < tmp.dim; i++)
		tmp.exp[i] = max(x.exp[i], y.exp[i]);
	return tmp;
};

monom gcd(monom &x, monom &y) {
	monom tmp(x);
	tmp.dim = min(x.dim, y.dim);
	for(int i = 0; i < tmp.dim; i++)
		tmp.exp[i] = min(x.exp[i], y.exp[i]);
	return tmp;
};

void prj(monom &x) { // proj, aff
	x.dim = proj;
	x.exp[proj - 1] = 0;
};
